@@ -9,6 +9,7 @@ import asyncio
 import socket
 import argparse
 import asyncssh
+import ipaddress
 import subprocess
 from operator import itemgetter
 
@@ -22,14 +23,29 @@ from operator import itemgetter
 #)
 
 image = "image.img"
-SSH_KEY_FILE = './ssh/id_rsa.key'
+SSH_KEY_FILE = 'id_rsa.key'
 SSH_PUBKEY_FILE = SSH_KEY_FILE + '.pub'
-NEXT_NODE_ADDR = 'fdca:ffee:8::1'
-SITE_LOCAL_PREFIX = 'fdca:ffee:8:0'
 HOST_ID = 1
 USE_CLIENT_TAP = False
 USE_NETNS = False
 
+
+# Special thanks to:
+#
+# https://github.com/NixOS/nixpkgs/blob/2577ec293255cbb995e42a86169cc40c427a6e7d/nixos/lib/test-driver/test-driver.py#L129-L140
+#
+def retry(fn) -> None:
+    """Call the given function repeatedly, with 1 second intervals,
+    until it returns True or a timeout is reached.
+    """
+    for _ in range(180):
+        if fn(False):
+            return
+        time.sleep(1)
+
+    if not fn(True):
+        raise Exception("action timed out")
+
 class Node():
 
     max_id = 0
@@ -45,10 +61,10 @@ class Node():
         self.if_index_max = 1
         self.uci_sets = []
         self.uci_commits = []
-        self.site_local_prefix = SITE_LOCAL_PREFIX
-        self.next_node_addr = NEXT_NODE_ADDR
         self.domain_code = None
         self.configured = False
+        self.addresses = []
+        self.dbg = debug_print(initial_time, self.hostname)
 
     def add_mesh_link(self, peer, _is_peer=False, _port=None):
         self.if_index_max += 1
@@ -86,6 +102,69 @@ class Node():
     def if_client(self):
         return "client" + str(self.id)
 
+    def execute_in_background(self, cmd, _msg=True):
+        class bg_cmd:
+            def __init__(self, node, cmd):
+                self.cmd = cmd
+                self.node = node
+                self.process = None
+
+                async def _async_execute():
+                    async with Node.ssh_conn(self.node) as conn:
+                        res = await conn.create_process(cmd, stderr=asyncssh.STDOUT)
+                        self.process = res
+                        return await res.wait()
+
+                self.task = loop.create_task(_async_execute())
+
+                if _msg:
+                    node.dbg(f'Command "{cmd}" started in background.')
+
+            def cancel(self):
+                self.process.send_signal('INT')
+                loop.run_until_complete(self.task)
+                self.node.dbg(f'Sent SIGINT to command "{self.cmd}".')
+
+        return bg_cmd(self, cmd)
+
+    def execute(self, cmd):
+        t = self.execute_in_background(cmd, _msg=False).task
+        loop.run_until_complete(t)
+
+        res = t.result()
+        return (res.exit_status, res.stdout.strip())
+
+    def succeed(self, cmd):
+        status, stdout = self.execute(cmd)
+
+        if status != 0:
+            msg = f'Expected success: command "{cmd}" failed with exit status {status}.'
+            self.dbg(msg)
+            self.dbg('Stdout/stderr was:')
+            for line in stdout.split('\n'):
+                self.dbg('| ' + line)
+            raise Exception(msg)
+        else:
+            self.dbg(f'Expected success: command "{cmd}" succeeded with exit status {status}.')
+            return stdout
+
+    def wait_until_succeeds(self, cmd):
+        output = ""
+
+        def check_success(is_last_attempt) -> bool:
+            nonlocal output
+            if is_last_attempt:
+                output = self.succeed(cmd)
+                return True
+            else:
+                status, output = self.execute(cmd)
+                return status == 0
+
+        self.dbg(f'Waiting until "{cmd}" succeeds.')
+        retry(check_success)
+        self.dbg(f'"{cmd}" succeeded.')
+        return output
+
     class ssh_conn:
 
         def __init__(self, node):
@@ -105,7 +184,8 @@ class Node():
                 if self.node.configured:
                     port += 100
 
-            conn = lambda: asyncssh.connect(addr, username='root', port=port, known_hosts=None, client_keys=[SSH_KEY_FILE])
+            keyfile = os.path.join(workdir, 'ssh', SSH_KEY_FILE)
+            conn = lambda: asyncssh.connect(addr, username='root', port=port, known_hosts=None, client_keys=[keyfile])
 
             # 100 retries
             for i in range(100):
@@ -166,10 +246,16 @@ def run_in_netns(netns, cmd):
 stdout_buffers = {}
 processes = {}
 masters = {}
+workdir = "./"
 
 async def gen_qemu_call(image, node):
 
-    shutil.copyfile('./' + image, './images/%02x.img' % node.id)
+    imgdir = os.path.join(workdir, 'images')
+    if not os.path.exists(imgdir):
+        os.mkdir(imgdir)
+
+    imgfile = os.path.join(imgdir, '%02x.img' % node.id)
+    shutil.copyfile('./' + image, imgfile)
 
     # TODO: machine identifier
     host_id = HOST_ID
@@ -209,6 +295,7 @@ async def gen_qemu_call(image, node):
     if USE_CLIENT_TAP:
         client_netdev = 'tap,id=hn2,script=no,downscript=no,ifname=%s' % node.if_client
     else:
+        # in config mode, the device is used for configuration with net 192.168.1.0/24
         client_netdev = 'user,id=hn2,hostfwd=tcp::' + str(ssh_port) + '-192.168.1.1:22,net=192.168.1.15/24'
 
     call = ['-nographic',
@@ -222,10 +309,13 @@ async def gen_qemu_call(image, node):
 
     # '-d', 'guest_errors', '-d', 'cpu_reset', '-gdb', 'tcp::' + str(3000 + node.id),
     args = ['qemu-system-x86_64',
-            '-drive', 'format=raw,file=./images/%02x.img' % node.id] + call + mesh_ifaces
+            '-drive', 'format=raw,file=' + imgfile] + call + mesh_ifaces
 
     master, slave = os.openpty()
-    pty_path = './ptys/node%d' % node.id;
+    ptydir = os.path.join(workdir, 'ptys')
+    if not os.path.exists(ptydir):
+        os.mkdir(ptydir)
+    pty_path = os.path.join(ptydir, 'node%d' % node.id)
     if os.path.islink(pty_path):
         os.remove(pty_path)
     os.symlink(os.ttyname(slave), pty_path)
@@ -254,8 +344,8 @@ async def set_mesh_devs(p, devs):
     await ssh_call(p, 'uci commit firewall')
 
 async def add_ssh_key(p):
-    # TODO: this removes baked in ssh keys :/
-    with open(SSH_PUBKEY_FILE) as f:
+    keyfile = os.path.join(workdir, 'ssh', SSH_PUBKEY_FILE)
+    with open(keyfile) as f:
         content = f.read()
         await ssh_call(p, 'cat >> /etc/dropbear/authorized_keys <<EOF\n' + content)
 
@@ -314,11 +404,22 @@ async def configure_node(initial_time, node):
     dbg('configuring node')
 
     async with Node.ssh_conn(node) as conn:
+        dbg('connection established')
         await config_node(initial_time, node, conn)
 
     dbg(node.hostname + ' configured')
     node.configured = True
 
+    # wait till all nodes are configured
+    for n in Node.all_nodes:
+        while not n.configured:
+            await asyncio.sleep(1)
+
+    # add /etc/hosts entries
+    async with Node.ssh_conn(node) as conn:
+        await add_hosts(conn)
+        dbg('/etc/hosts is now adjusted')
+
     if USE_CLIENT_TAP and USE_NETNS:
         configure_netns(node)
 
@@ -346,7 +447,12 @@ def read_to_buffer(node):
     process = processes[node.id]
     master = masters[node.id]
     stdout_buffers[node.id] = b""
-    with open('logs/' + node.hostname + '.log', 'wb') as f1:
+
+    logdir = os.path.join(workdir, 'logs')
+    if not os.path.exists(logdir):
+        os.mkdir(logdir)
+
+    with open(os.path.join(logdir, node.hostname + '.log'), 'wb') as f1:
         while True:
             b = yield from process.stdout.read(1) # TODO: is this unbuffered?
             stdout_buffers[node.id] += b
@@ -370,6 +476,12 @@ def wait_for(node, b):
         yield from asyncio.sleep(0)
 
 async def add_hosts(p):
+    host_entries = ""
+
+    for n in Node.all_nodes:
+        for a in n.addresses:
+            host_entries += str(a) + " " + n.hostname + "\n"
+
     await ssh_call(p, 'cat >> /etc/hosts <<EOF\n' + host_entries + '\n')
     await ssh_call(p, 'cat >> /etc/bat-hosts <<EOF\n' + bathost_entries + '\n')
 
@@ -388,7 +500,6 @@ async def config_node(initial_time, node, ssh_conn):
     mesh_ifaces = list(map(itemgetter(0), node.mesh_links))
 
     await set_mesh_devs(p, mesh_ifaces)
-    await add_hosts(p)
     await ssh_call(p, 'pretty-hostname ' + node.hostname)
     await add_ssh_key(p)
 
@@ -401,6 +512,12 @@ async def config_node(initial_time, node, ssh_conn):
     if node.domain_code is not None:
         await ssh_call(p, "gluon-reconfigure")
 
+    prefix = (await ssh_call(p, 'gluon-show-site | jsonfilter -e @.prefix6')).strip()
+    prefix = ipaddress.ip_network(prefix)
+
+    mac = (await ssh_call(p, 'uci get network.client.macaddr')).strip()
+    node.addresses.append(mac_to_ip6(mac, prefix))
+
     # reboot to operational mode
     await ssh_call(p, 'uci set gluon-setup-mode.@setup_mode[0].configured=\'1\'')
     await ssh_call(p, 'uci commit gluon-setup-mode')
@@ -440,7 +557,14 @@ loop = None
 config_tasks = []
 args = None
 
+def mac_to_ip6(mac, net):
+    mac = list(map(lambda x: int(x, 16), mac.split(':')))
+    x = list(next(net.hosts()).packed)
+    x[8:] = [mac[0] ^ 0x02] + mac[1:3] + [0xff, 0xfe] + mac[3:]
+    return ipaddress.ip_address(bytes(x))
+
 def start():
+    global workdir
     global configured
     if configured:
         return
@@ -449,13 +573,24 @@ def start():
     parser = argparse.ArgumentParser()
     parser.add_argument("--run-forever", help="", action="store_true")
     parser.add_argument("--run-tests-on-existing-instance", help="", action="store_true")
+    parser.add_argument("--use-tmp-workdir", help="", action="store_true")
     args = parser.parse_args()
 
+    if args.use_tmp_workdir:
+        workdir = os.path.join('/tmp', 'gluon-qemu-testlab')
+
+        if not os.path.exists(workdir):
+            os.mkdir(workdir)
+
     #if os.environ.get('TMUX') is None and not 'notmux' in sys.argv:
     #    os.execl('/usr/bin/tmux', 'tmux', '-S', 'test', 'new', sys.executable, '-i', *sys.argv)
 
-    if not os.path.exists(SSH_PUBKEY_FILE):
-        run('ssh-keygen -t rsa -f ' + SSH_KEY_FILE + ' -N \'\'')
+    sshdir = os.path.join(workdir, 'ssh')
+    if not os.path.exists(sshdir):
+        os.mkdir(sshdir)
+
+    if not os.path.exists(os.path.join(sshdir, SSH_PUBKEY_FILE)):
+        run('ssh-keygen -t rsa -f ' + os.path.join(sshdir, SSH_KEY_FILE) + ' -N \'\'')
 
     global loop
     loop = asyncio.get_event_loop()
@@ -473,11 +608,7 @@ def start():
     global config_tasks
 
     for node in Node.all_nodes:
-        host_entries += "{node.site_local_prefix}:5054:{host_id}ff:fe{node.id:02x}:3402 {node.hostname}\n".format(node=node, host_id=host_id)
-        client_name = node.hostname.replace('node', 'client')
-        host_entries += "{node.site_local_prefix}:a854:{host_id}ff:fe{node.id:02x}:3402 {client_name}\n".format(node=node, host_id=host_id, client_name=client_name)
         bathost_entries += "52:54:{host_id:02x}:{node.id:02x}:34:02 {node.hostname}\n".format(node=node, host_id=host_id)
-        bathost_entries += "aa:54:{host_id:02x}:{node.id:02x}:34:02 {client_name}\n".format(node=node, host_id=host_id, client_name=client_name)
 
     bathost_entries += "de:ad:be:ee:ff:01 mobile1\n"
 
@@ -507,119 +638,6 @@ def finish():
 def connect(a, b):
     a.add_mesh_link(b)
 
-####
-#### TESTING API
-####
-
-p_id = 1
-ssh_processes = {}
-async def _ssh(p):
-    n = p['node']
-    cmd = p['cmd']
-    global ssh_processes
-    async with Node.ssh_conn(n) as c:
-        res = await c.create_process(cmd)
-        p['process'] = res
-        return await res.wait()
-
-def ssh(node, cmd):
-    global p_id
-    global ssh_processes
-
-    p = { "node": node, "process": None, "task": None, "exit_with_others": False, "cmd": cmd, "expect_success": False }
-    ssh_processes[p_id] = p
-    p_id += 1
-
-    return p
-
-def ssh_singlecmd(node, cmd):
-    ssh(node, cmd)
-    sync()
-
-def exit_with_others(p):
-    p['exit_with_others'] = True
-    return p
-
-def expect_success(p):
-    p['expect_success'] = True
-    return p
-
-def _sync():
-    global ssh_processes
-
-    for proc in ssh_processes.values():
-        proc['task'] = loop.create_task(_ssh(proc))
-
-    # we wait here until at least all ssh processes are started
-    for proc in ssh_processes.values():
-        retries = 100
-        while True:
-            if proc['process'] is None:
-                # this happens if c.create_process() has not yet been awaited
-                print('The process "' + proc['cmd'] + '" has not started yet. ' + str(retries) + ' retries left.')
-                t = loop.create_task(asyncio.sleep(1))
-                loop.run_until_complete(t)
-
-                if retries < 1:
-                    exit(1)
-
-                retries -= 1
-            else:
-                break
-
-    for proc in ssh_processes.values():
-        if proc['exit_with_others']:
-            continue
-
-        loop.run_until_complete(proc['task'])
-
-    for proc in ssh_processes.values():
-        p = proc['process']
-        t = proc['task']
-
-        if p.exit_status is None:
-            print('sending SIGINT to "' + p.command + '"')
-            p.send_signal('INT')
-        else:
-            print('command "' + p.command + '" exited with status code ' + str(p.exit_status))
-            print('stdout:')
-            print(t.result().stdout)
-            print('stderr:')
-            print(t.result().stderr)
-
-    success = True
-    for proc in ssh_processes.values():
-        if proc['expect_success'] and proc['process'].exit_status > 0:
-            success = False
-
-    if success:
-        ssh_processes = {}
-
-    return success
-
-def sync(retries=1, sleep=5):
-    while True:
-        success = _sync()
-        if success:
-            break
-
-        retries -= 1
-
-        if retries < 1:
-            print('TESTS FAILED!')
-            exit(1)
-
-        print('retrying. ' + str(retries) + ' retries left. retrying in ' + str(sleep) + ' seconds.')
-        time.sleep(sleep)
-        print('retrying now.')
-
-def check(p):
-    global ssh_processes
-    expect_success(p)
-    success = _sync()
-    ssh_processes = {}
-    return success
-
 def new_loop():
     global loop
     loop = asyncio.get_event_loop()
